from ..utilities.util import Util
from ..trees.event import EventTree
from fractions import Fraction
from operator import add, sub
from IPython.display import Image
from IPython import get_ipython
import networkx as nx
import scipy.special
import logging

logger = logging.getLogger('cegpy.staged_tree')


class StagedTree(EventTree):
    def __init__(
            self,
            dataframe,
            prior=None,
            alpha=None,
            hyperstage=None,
            sampling_zero_paths=None,
            incoming_graph_data=None,
            **attr) -> None:

        # Call event tree init to generate event tree
        super().__init__(
            dataframe,
            sampling_zero_paths,
            incoming_graph_data,
            **attr
        )

        self.__store_params(prior, alpha, hyperstage)
        self._mean_posterior_probs = []
        self._merged_situations = []
        self._stage_colours = []
        self._sort_count = 0
        self._colours_for_situations = []
        logger.debug("Starting Staged Tree")

    @property
    def prior(self):
        return nx.get_edge_attributes(self, 'prior')

    @prior.setter
    def prior(self, prior):
        offset = 0
        for node_idx, node_priors in enumerate(prior):
            node_name = ('s%d' % (node_idx + offset))
            while self.succ[node_name] == {}:
                offset += 1
                node_name = ('s%d' % (node_idx + offset))
            pass
            for edge_prior_idx, succ_key in enumerate(
                    self.succ[node_name].keys()):
                label = list(self.succ[node_name][succ_key])[0]
                self.edges[(node_name, succ_key, label)]['prior'] = \
                    node_priors[edge_prior_idx]

    def get_prior_as_list(self):
        prior_list = []
        prev_node = list(self.prior)[0][0]
        succ_list = []

        for edge, prior in self.prior.items():
            node = edge[0]
            if node != prev_node:
                prior_list.append(succ_list)
                succ_list = []
            succ_list.append(prior)
            prev_node = node

        if succ_list != []:
            prior_list.append(succ_list)
        return prior_list

    @property
    def posterior(self):
        '''Posterior is calculated such that the edge count is added
        to the prior for each edge.'''
        try:
            posterior = nx.get_edge_attributes(self, 'posterior')
            if posterior == {}:
                raise AttributeError('Posterior not yet set.')
            else:
                return posterior
        except AttributeError:
            for edge in self.edges:
                edge_dict = self.edges[edge]
                posterior = edge_dict['prior'] + edge_dict['count']
                edge_dict['posterior'] = posterior
            return nx.get_edge_attributes(self, 'posterior')

    def get_posterior_as_list(self):
        posterior_list = []
        prev_node = list(self.posterior)[0][0]
        succ_list = []

        for edge, posterior in self.posterior.items():
            node = edge[0]
            if node != prev_node:
                posterior_list.append(succ_list)
                succ_list = []
            succ_list.append(posterior)
            prev_node = node

        if succ_list != []:
            posterior_list.append(succ_list)
        return posterior_list

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self._alpha = value

    @property
    def hyperstage(self):
        return self._hyperstage

    @hyperstage.setter
    def hyperstage(self, value):
        self._hyperstage = value

    @property
    def edge_countset(self):
        return self.__create_edge_countset()

    @property
    def ahc_output(self):
        return self._ahc_output

    @ahc_output.setter
    def ahc_output(self, value):
        self._ahc_output = value

    def __store_params(self, prior, alpha, hyperstage) -> None:
        """User has passed in AHC params, this function processes them,
        and generates any default AHC params if required."""
        if prior:
            if alpha:
                self.alpha = None
                logging.warning("Params Warning!! When prior is given, " +
                                "alpha is not required!")
        else:
            if alpha is None:
                self.alpha = self.__calculate_default_alpha()
                logging.warning("Params Warning!! Neither prior nor alpha " +
                                "were provided. Using default alpha " +
                                "value of %d.", self.alpha)
            else:
                self.alpha = alpha

            # No matter what alpha is, generate default prior
            self.prior = self.__create_default_prior(self.alpha)

        if hyperstage is None:
            self.hyperstage = self.__create_default_hyperstage()
        else:
            self.hyperstage = hyperstage

    def __calculate_default_alpha(self) -> int:
        """If no alpha is given, a default value is calculated.
        The value is calculated by determining the maximum number
        of categories that any one variable has"""
        logger.info("Calculating default prior")
        max_count = max(list(self.categories_per_variable.values()))
        return max_count

    def __create_default_prior(self, alpha) -> list:
        """default prior set for the AHC method using the mass conservation property.
        That is, the alpha param is the phantom sample starting at
        the root, and it is spread equally across all edges along the tree.
        (see chapter 5 of Collazo, Gorgen & Smith 'Chain Event Graphs', 2018)
        The prior is a list of lists. Each list gives the prior along the
        edges of a specific situation.
        Indexed same as self.situations & self.egde_countset"""

        logger.info("Generating default prior")
        default_prior = [0] * len(self.situations)
        sample_size_at_node = dict()

        # Root node is assigned phantom sample (alpha)
        if isinstance(alpha, float):
            alpha = Fraction.from_float(alpha)
        elif isinstance(alpha, int) or isinstance(alpha, str):
            alpha = Fraction.from_float(float(alpha))
        else:
            logger.warning("Prior generator param alpha is in a strange format.\
                            ..")

        sample_size_at_node[self.root] = alpha

        for node_idx, node in enumerate(self.situations):
            # How many nodes emanate from the current node?
            number_of_emanating_nodes = self.out_degree[node]

            # Divide the sample size from the current node equally among
            # emanating nodes
            equal_distribution_of_sample = \
                sample_size_at_node[node] / number_of_emanating_nodes
            default_prior[node_idx] = \
                [equal_distribution_of_sample] * number_of_emanating_nodes

            relevant_terminating_nodes = [
                edge[1] for edge in list(self.edges) if edge[0] == node
            ]

            for terminating_node in relevant_terminating_nodes:
                sample_size_at_node[terminating_node] = \
                 equal_distribution_of_sample

        return default_prior

    def __create_default_hyperstage(self) -> list:
        '''Generates default hyperstage for the AHC method.
        A hyperstage is a list of lists such that two situaions can be in the
        same stage only if there are elements of the same list for some list
        in the hyperstage.
        The default is to allow all situations with the same number of
        outgoing edges and the same edge labels to be in a common list. '''
        logger.info("Creating default hyperstage")
        hyperstage = []
        info_of_edges = []

        for node in self.situations:
            labels = [
                edge[2] for edge in self.edges
                if edge[0] == node
            ]
            labels.sort()

            info_of_edges.append([self.out_degree[node], labels])

        sorted_info = []
        for info in info_of_edges:
            if info not in sorted_info:
                sorted_info.append(info)

        for info in sorted_info:
            situations_with_value_edges = []
            for idx, situation in enumerate(self.situations):
                if info_of_edges[idx] == info:
                    situations_with_value_edges.append(situation)
            hyperstage = hyperstage + [situations_with_value_edges]

        return hyperstage

    def __create_edge_countset(self) -> list:
        '''Each element of list contains a list with counts along edges emanating from
        a specific situation. Indexed same as self.situations'''
        logger.info("Creating edge countset")
        edge_countset = []

        for node in self.situations:
            edge_countset.append([
                count for edge, count in self.edge_counts.items()
                if edge[0] == node
            ])
        return edge_countset

    def _calculate_lg_of_sum(self, array) -> float:
        '''function to calculate log gamma of the sum of an array'''
        array = [float(x) for x in array]
        return scipy.special.gammaln(sum(array))

    def _calculate_sum_of_lg(self, array) -> float:
        '''function to calculate log gamma of each element of an array'''
        return sum([scipy.special.gammaln(float(x)) for x in array])

    def _calculate_initial_loglikelihood(self, prior, posterior) -> float:
        '''calculating log likelihood given a prior and posterior'''
        # Calculate prior contribution
        logger.info("Calculating initial loglikelihood")

        pri_lg_of_sum = [
            self._calculate_lg_of_sum(elem) for elem in prior
        ]
        pri_sum_of_lg = [
            self._calculate_sum_of_lg(elem) for elem in prior
        ]
        pri_contribution = list(map(sub, pri_lg_of_sum, pri_sum_of_lg))

        # Calculate posterior contribution
        post_lg_of_sum = [
            self._calculate_lg_of_sum(elem) for elem in posterior
        ]
        post_sum_of_lg = [
            self._calculate_sum_of_lg(elem) for elem in posterior
        ]
        post_contribution = list(map(sub, post_sum_of_lg, post_lg_of_sum))

        sum_pri_contribution = sum(pri_contribution)
        sum_post_contribution = sum(post_contribution)
        return (sum_pri_contribution + sum_post_contribution)

    def _check_issubset(self, item, hyperstage) -> bool:
        '''function to check if two situations belong to the same set in the
        hyperstage'''
        return any(set(item).issubset(element) for element in hyperstage)

    def _calculate_bayes_factor(self, prior1, posterior1,
                                prior2, posterior2) -> float:
        '''calculates the bayes factor comparing two models which differ in
        only one stage'''
        new_prior = list(map(add, prior1, prior2))
        new_posterior = list(map(add, posterior1, posterior2))
        return (
            self._calculate_lg_of_sum(new_prior)
            - self._calculate_lg_of_sum(prior1)
            - self._calculate_lg_of_sum(prior2)
            - self._calculate_lg_of_sum(new_posterior)
            + self._calculate_lg_of_sum(posterior1)
            + self._calculate_lg_of_sum(posterior2)
            + self._calculate_sum_of_lg(new_posterior)
            - self._calculate_sum_of_lg(posterior1)
            - self._calculate_sum_of_lg(posterior2)
            - self._calculate_sum_of_lg(new_prior)
            + self._calculate_sum_of_lg(prior1)
            + self._calculate_sum_of_lg(prior2)
        )

    def _sort_list(self, a_list_of_lists) -> list:
        '''function to sort a list of lists to remove repetitions'''

        for l1_idx in range(0, len(a_list_of_lists)):
            for l2_idx in range(l1_idx+1, len(a_list_of_lists)):
                list_1 = a_list_of_lists[l1_idx]
                list_2 = a_list_of_lists[l2_idx]
                lists_intersect = set(list_1) & set(list_2)

                if lists_intersect:
                    lists_union = list(set(list_1) | set(list_2))
                    # new_list_of_lists.append(lists_union)
                    a_list_of_lists[l1_idx] = []
                    a_list_of_lists[l2_idx] = lists_union

        new_list_of_lists = [
            elem for elem in a_list_of_lists
            if elem != []
        ]

        if new_list_of_lists == a_list_of_lists:
            return new_list_of_lists
        else:
            return self._sort_list(new_list_of_lists)

    def _calculate_mean_posterior_probs(self, probs) -> list:
        '''Iterates through array of lists, calculates mean
        posterior probabilities'''
        mean_posterior_probs = []
        for arr in probs:
            total = sum(arr)
            mean_posterior_probs.append(
                [round(element/total, 3) for element in arr]
            )
        return mean_posterior_probs

    def _execute_AHC_algoritm(self):
        # prior = self.prior.copy()
        hyperstage = self.hyperstage.copy()
        # posterior = self.posterior.copy()
        prior_list = self.get_prior_as_list().copy()
        posterior_list = self.get_posterior_as_list().copy()
        loglikelihood = self._calculate_initial_loglikelihood(
            prior_list, posterior_list
        )
        posterior_probs = self.get_posterior_as_list().copy()
        situ = self.situations.copy()
        merged_situation_list = []
        bayesfactor_score = 1

        logger.info(" ----- Starting main loop of AHC algorithm -----")
        logger.info("Prior is length %d" % len(prior_list))
        while bayesfactor_score > 0:
            local = {}

            for sit_1 in range(len(prior_list)):
                if all(items == 0 for items in posterior_list[sit_1]) is False:
                    model1 = [prior_list[sit_1], posterior_list[sit_1]]
                    for sit_2 in range(sit_1+1, len(prior_list)):
                        is_subset = self._check_issubset(
                            [situ[sit_1], situ[sit_2]],
                            hyperstage
                        )

                        any_non_zero = all(
                            items == 0 for items in posterior_list[sit_2]
                        )

                        if is_subset and not any_non_zero:
                            model2 = [prior_list[sit_2], posterior_list[sit_2]]
                            local[(sit_1, sit_2)] = \
                                self._calculate_bayes_factor(
                                    *model1, *model2
                                )

            max_key = max(local, key=local.get)
            bayesfactor_score = local[max_key]

            if local != {} and bayesfactor_score > 0:
                merged_situation_list.append(list(max_key))

                prior_list[max_key[0]] = list(
                    map(
                        add,
                        prior_list[max_key[0]],
                        prior_list[max_key[1]]
                    )
                )
                posterior_list[max_key[0]] = list(
                    map(
                        add,
                        posterior_list[max_key[0]],
                        posterior_list[max_key[1]]
                    )
                )

                prior_list[max_key[1]] = \
                    [0] * len(prior_list[max_key[0]])
                posterior_list[max_key[1]] = \
                    [0] * len(prior_list[max_key[0]])

                posterior_probs[max_key[0]] = posterior_list[max_key[0]]
                posterior_probs[max_key[1]] = posterior_list[max_key[0]]

                loglikelihood += bayesfactor_score

            logger.debug('--Current AHC score: %d' % bayesfactor_score)

        return posterior_probs, loglikelihood, merged_situation_list

    def _mark_nodes_with_stage_number(self, merged_situation_indexes) -> list:
        """AHC algorithm creates a list of indexes to the situations list.
        This function takes those indexes and creates a new list which is
        in a string representation of nodes."""
        self._sort_count = 0
        list_of_merged_situations = self._sort_list(merged_situation_indexes)
        for index, stage in enumerate(list_of_merged_situations):
            for node_number in stage:
                node_name = 's' + str(node_number)
                self.nodes[node_name]['stage'] = index

        return list_of_merged_situations

    def _generate_colours_for_situations(self, merged_situations):
        """Colours each stage of the tree with an individual colour"""
        number_of_stages = len(merged_situations)
        stage_colours = Util.generate_colours(number_of_stages)
        self._stage_colours = stage_colours

        for node in self.nodes:
            try:
                stage = self.nodes[node]['stage']
                self.nodes[node]['colour'] = stage_colours[stage]
            except KeyError:
                self.nodes[node]['colour'] = 'lightgrey'

    def calculate_AHC_transitions(self, prior=None,
                                  alpha=None, hyperstage=None):
        '''Bayesian Agglommerative Hierarchical Clustering algorithm
        implementation. It returns a list of lists of the situations which
        have been merged together, the likelihood of the final model and
        the mean posterior conditional probabilities of the stages.'''
        logger.info("\n\n --- Starting AHC Algorithm ---")

        self.__store_params(prior, alpha, hyperstage)

        posterior_probs, loglikelihood, \
            merged_situation_indexes = self._execute_AHC_algoritm()

        mean_posterior_probs = \
            self._calculate_mean_posterior_probs(posterior_probs)

        merged_situations = \
            self._mark_nodes_with_stage_number(merged_situation_indexes)

        self._generate_colours_for_situations(merged_situations)

        self.ahc_output = {
            "Merged Situations": merged_situations,
            "Loglikelihood": loglikelihood,
            "Mean Posterior Probabilities": mean_posterior_probs
        }
        return self.ahc_output

    def create_figure(self, filename):
        """Draws the coloured staged tree for the process described by
        the dataset, and saves it to "<filename>.filetype". Supports
        any filetype that graphviz supports. e.g: "event_tree.png" or
        "event_tree.svg" etc.
        """
        try:
            self.ahc_output
            filename, filetype = Util.generate_filename_and_mkdir(filename)
            logger.info("--- generating graph ---")
            graph = self._generate_pdp_graph()
            logger.info("--- writing " + filetype + " file ---")
            graph.write(str(filename), format=filetype)

            if get_ipython() is None:
                return None
            else:
                logger.info("--- Exporting graph to notebook ---")
                return Image(graph.create_png())

        except AttributeError:
            logger.error(
                "----- PLEASE RUN AHC ALGORITHM before trying to" +
                " export graph -----"
            )
            return None



import pydotplus as pdp
import networkx as nx
from copy import deepcopy
import itertools as it

from ..utilities.util import Util
from IPython.display import Image
from IPython import get_ipython
import logging

logger = logging.getLogger('cegpy.chain_event_graph')


class ChainEventGraph(nx.MultiDiGraph):
    """
    Class: Chain Event Graph

    Input: Staged tree object (StagedTree)
    Output: Chain event graphs
    """
    def __init__(self, staged_tree=None, **attr):
        super().__init__(staged_tree, **attr)
        self.sink_suffix = '&infin;'
        self.node_prefix = 'w'

        if staged_tree is not None:
            try:
                self.ahc_output = deepcopy(staged_tree.ahc_output)
            except AttributeError:
                self.ahc_output = {}
        else:
            logger.info("Class called with no incoming graph.")
        self.evidence = Evidence(self)

    @property
    def node_prefix(self):
        return self._node_prefix

    @node_prefix.setter
    def node_prefix(self, value):
        self._node_prefix = str(value)

    @property
    def sink_suffix(self):
        return self._sink_suffix

    @sink_suffix.setter
    def sink_suffix(self, value):
        self._sink_suffix = str(value)

    @property
    def sink_node(self):
        return "%s%s" % (self.node_prefix, self.sink_suffix)

    @property
    def root_node(self):
        return ("%s0" % self.node_prefix)

    @property
    def certain_evidence(self):
        return self._certain_evidence

    @certain_evidence.setter
    def certain_evidence(self, value):
        self._certain_evidence = value

    @property
    def uncertain_evidence(self):
        return self._uncertain_evidence

    @uncertain_evidence.setter
    def uncertain_evidence(self, value):
        self._uncertain_evidence = value

    @property
    def path_list(self):
        return self.__path_list

    @path_list.setter
    def path_list(self, value):
        self.__path_list = value

    @property
    def reduced(self):
        return self.evidence.reduced_graph

    @property
    def stages(self):
        self.__stages = {}
        node_stages = dict(self.nodes(data='stage', default=None))
        for k, v in node_stages.items():
            try:
                self.__stages[v].append(k)
            except KeyError:
                self.__stages[v] = [k]

        return self.__stages

    def clear_evidence(self):
        self.evidence = Evidence(self)

    def generate(self):
        '''
        This function takes the output of the AHC algorithm and identifies
        the positions i.e. the vertices of the CEG and the edges of the CEG
        along with their edge labels and edge counts. Here we use the
        algorithm in our paper with the optimal stopping time.
        '''
        def check_vertices_can_be_merged(v1, v2) -> bool:
            has_same_successor_nodes = \
                set(self.adj[v1].keys()) == set(self.adj[v2].keys())

            if has_same_successor_nodes:
                has_same_outgoing_edges = True
                v1_adj = self.succ[v1]
                for succ_node in list(v1_adj.keys()):
                    v1_edges = self.succ[v1][succ_node]
                    v2_edges = self.succ[v2][succ_node]

                    if v1_edges is None or v2_edges is None:
                        has_same_outgoing_edges &= False
                        break

                    v2_edge_labels = \
                        [label for label in v2_edges.keys()]

                    for label in v1_edges.keys():
                        if label not in v2_edge_labels:
                            has_same_outgoing_edges &= False
                            break
                        else:
                            has_same_outgoing_edges &= True
            else:
                has_same_outgoing_edges = False

            try:
                in_same_stage = \
                    self.nodes[v1]['stage'] == self.nodes[v2]['stage']
            except KeyError:
                in_same_stage = False

            return in_same_stage and \
                has_same_successor_nodes and has_same_outgoing_edges

        def merge_nodes(nodes_to_merge):
            """nodes to merge should be a set of 2 element tuples"""
            temp_1 = 'temp_1'
            temp_2 = 'temp_2'
            while nodes_to_merge != set():
                nodes = nodes_to_merge.pop()
                new_node = nodes[0]
                # Copy nodes to temp nodes
                node_map = {
                    nodes[0]: temp_1,
                    nodes[1]: temp_2
                }
                nx.relabel_nodes(self, node_map, copy=False)
                ebunch_to_remove = []  # List of edges to remove
                self.add_node(new_node)
                for succ, t1_edge_dict in self.succ[temp_1].items():
                    edge_labels = list(t1_edge_dict.keys())
                    while edge_labels != []:
                        new_edge_data = {}
                        label = edge_labels.pop(0)
                        t1_edge = t1_edge_dict[label]
                        t2_edge = self.succ[temp_2][succ][label]
                        new_edge_data['count'] = \
                            t1_edge['count'] + t2_edge['count']
                        new_edge_data['prior'] = \
                            t1_edge['prior'] + t2_edge['prior']
                        new_edge_data['posterior'] = \
                            t1_edge['posterior'] + t2_edge['posterior']
                        try:

                            new_edge_data['probability'] = \
                                t1_edge['probability']
                            self.add_edge(
                                u_for_edge=new_node,
                                v_for_edge=succ,
                                key=label,
                                count=new_edge_data['count'],
                                prior=new_edge_data['prior'],
                                posterior=new_edge_data['posterior'],
                                probability=new_edge_data['probability']
                            )
                        except KeyError:
                            self.add_edge(
                                u_for_edge=new_node,
                                v_for_edge=succ,
                                key=label,
                                count=new_edge_data['count'],
                                prior=new_edge_data['prior'],
                                posterior=new_edge_data['posterior']
                            )
                        ebunch_to_remove.append((temp_1, succ, label))
                        ebunch_to_remove.append((temp_2, succ, label))

                self.remove_edges_from(ebunch_to_remove)
                nx.relabel_nodes(
                    G=self,
                    mapping={temp_1: new_node, temp_2: new_node},
                    copy=False
                )

                # Some nodes have been removed, we need to update the
                # mergeable list to point to new nodes if required
                temp_list = list(nodes_to_merge)
                for pair in temp_list:
                    if nodes[1] in pair:
                        new_pair = (
                            # the other node of the pair
                            pair[pair.index(nodes[1]) - 1],
                            # the new node it will be merged to
                            new_node
                        )
                        nodes_to_merge.remove(pair)
                        if new_pair[0] != new_pair[1]:
                            nodes_to_merge.add(new_pair)
                pass

        def relabel_nodes(base_nodes, renamed_nodes=[]):
            next_level = []
            # first, relabel the successors of this node
            for node in base_nodes:
                node_mapping = {}
                for succ in self.succ[node].keys():
                    if succ != self.sink_node and succ not in renamed_nodes:
                        node_mapping[succ] = self.__get_next_node_name()
                        next_level.append(node_mapping[succ])
                        renamed_nodes.append(node_mapping[succ])

                if node_mapping != {}:
                    nx.relabel_nodes(
                        self,
                        node_mapping,
                        copy=False
                    )
            if next_level != []:
                relabel_nodes(next_level, renamed_nodes)

        if self.ahc_output == {}:
            raise ValueError("Run staged tree AHC transitions first.")
        # rename root node:
        nx.relabel_nodes(self, {'s0': self.root_node}, copy=False)
        self.__update_probabilities()
        self.__trim_leaves_from_graph()
        self.__update_distances_of_nodes_to_sink_node()
        src_node_gen = self.__gen_nodes_with_increasing_distance(
            start=1
        )
        next_set_of_nodes = next(src_node_gen)

        while next_set_of_nodes != [self.root_node]:
            nodes_to_merge = set()
            while len(next_set_of_nodes) > 1:
                node_1 = next_set_of_nodes.pop(0)
                for node_2 in next_set_of_nodes:
                    mergeable = check_vertices_can_be_merged(node_1, node_2)
                    if mergeable:
                        nodes_to_merge.add((node_1, node_2))

            merge_nodes(nodes_to_merge)

            try:
                next_set_of_nodes = next(src_node_gen)
            except StopIteration:
                next_set_of_nodes = []

        relabel_nodes([self.root_node])
        self.__update_path_list()

    def create_figure(self, filename):
        """
        Draws the chain event graph representation of the stage tree,
        and saves it to "<filename>.filetype". Supports any filetype that
        graphviz supports. e.g: "event_tree.png" or "event_tree.svg" etc.
        """
        filename, filetype = Util.generate_filename_and_mkdir(filename)
        graph = pdp.Dot(graph_type='digraph', rankdir='LR')
        edge_probabilities = list(
            self.edges(data='probability', default=1, keys=True)
        )

        for (u, v, k, p) in edge_probabilities:
            full_label = "{}\n{:.2f}".format(k, p)
            graph.add_edge(
                pdp.Edge(
                    src=u,
                    dst=v,
                    label=full_label,
                    labelfontcolor='#009933',
                    fontsize='10.0',
                    color='black'
                )
            )
        nodes = list(nx.topological_sort(self))
        for node in nodes:
            try:
                fill_colour = self.nodes[node]['colour']
            except KeyError:
                fill_colour = 'white'
            label = "<" + node[0] + "<SUB>" + node[1:] + "</SUB>" + ">"
            graph.add_node(
                pdp.Node(
                    name=node,
                    label=label,
                    style='filled',
                    fillcolor=fill_colour
                )
            )

        graph.write(str(filename), format=filetype)

        if get_ipython() is None:
            return None
        else:
            return Image(graph.create_png())

    def __update_probabilities(self):
        count_total_lbl = 'count_total'
        edge_counts = list(self.edges(data='count', keys=True, default=0))

        for stage, stage_nodes in self.stages.items():
            count_total = 0
            stage_edges = {}
            if stage is not None:

                for (u, _, k, c) in edge_counts:
                    if u in stage_nodes:
                        count_total += c
                        try:
                            stage_edges[k] += c
                        except KeyError:
                            stage_edges[k] = c

                for node in stage_nodes:
                    self.nodes[node][count_total_lbl] = count_total

                for (u, v, k, _) in edge_counts:
                    if u in stage_nodes:
                        self.edges[u, v, k]['probability'] =\
                            stage_edges[k] / count_total
            else:
                for node in stage_nodes:
                    count_total = 0
                    stage_edges = {}
                    for (u, _, k, c) in edge_counts:
                        if u == node:
                            count_total += c
                            try:
                                stage_edges[k] += c
                            except KeyError:
                                stage_edges[k] = c

                    self.nodes[node][count_total_lbl] = count_total
                    for (u, v, k, _) in edge_counts:
                        if u == node:
                            self.edges[u, v, k]['probability'] =\
                                stage_edges[k] / count_total

    def __update_path_list(self) -> None:
        path_generator = nx.all_simple_edge_paths(
            self,
            self.root_node,
            self.sink_node
        )
        path_list = []
        while True:
            try:
                path_list.append(next(path_generator))
            except StopIteration:
                self.path_list = path_list
                break

    def __update_distances_of_nodes_to_sink_node(self) -> None:
        """
        Iterates through the graph until it finds the root node.
        For each node, it determines the maximum number of edges
        from that node to the sink node.
        """
        max_dist = 'max_dist_to_sink'
        self.nodes[self.sink_node][max_dist] = 0
        node_queue = [self.sink_node]

        while node_queue != [self.root_node]:
            node = node_queue.pop(0)
            for pred in self.predecessors(node):
                max_dist_to_sink = set()
                for succ in self.successors(pred):
                    max_dist_to_sink.add(
                        self.nodes[succ][max_dist]
                    )

                self.nodes[pred][max_dist] = max(max_dist_to_sink) + 1
                if pred not in node_queue:
                    node_queue.append(pred)

    def __gen_nodes_with_increasing_distance(self, start=0) -> list:
        max_dists = nx.get_node_attributes(self, 'max_dist_to_sink')
        distance_dict = {}
        for key, value in max_dists.items():
            distance_dict.setdefault(value, []).append(key)

        for dist in range(len(distance_dict)):
            if dist >= start:
                yield distance_dict[dist]

    def __get_next_node_name(self):
        try:
            num = str(next(self._num_iter))
        except AttributeError:
            self._num_iter = it.count(1, 1)
            num = str(next(self._num_iter))

        return str(self.node_prefix) + num

    def __trim_leaves_from_graph(self):
        # Create new CEG sink node
        self.add_node(self.sink_node, colour='lightgrey')
        outgoing_edges = deepcopy(self.succ).items()
        # Check to see if any nodes have no outgoing edges.
        for node, outgoing_edges in outgoing_edges:
            if outgoing_edges == {} and node != self.sink_node:
                incoming_edges = deepcopy(self.pred[node]).items()
                # When node is identified as a leaf check the
                # predessesor nodes that have edges that enter this node.
                for pred_node, edges in incoming_edges:
                    for edge_label, edge in edges.items():
                        # Create new edge that points to the sink node,
                        # with all the same data as the edge we will delete.
                        try:
                            prob = edge['probability']
                        except KeyError:
                            prob = 1
                        self.add_edge(
                            pred_node,
                            self.sink_node,
                            key=edge_label,
                            count=edge['count'],
                            prior=edge['prior'],
                            posterior=edge['posterior'],
                            probability=prob
                        )
                self.remove_node(node)


class Evidence:
    CERTAIN = True
    UNCERTAIN = False

    def __init__(self, graph):
        self.__graph = graph

        self.certain_edges = []
        self.uncertain_edges = []
        self.certain_vertices = set()
        self.uncertain_vertices = set()

    @property
    def reduced_graph(self):
        return self.__create_reduced_graph()

    @property
    def path_list(self):
        return self._path_list

    @path_list.setter
    def path_list(self, value):
        self._path_list = value

    @property
    def edges(self):
        return list(self._edges)

    @edges.setter
    def edges(self, value):
        self._edges = value

    @property
    def vertices(self):
        return list(self._vertices)

    @vertices.setter
    def vertices(self, value):
        self._vertices = value

    def add_edge(self, u, v, label, certain):
        edge = (u, v, label)
        if certain:
            self.certain_edges.append(edge)
        else:
            self.uncertain_edges.append(edge)

    def add_edges_from(self, edges, certain):
        for (u, v, k) in edges:
            self.add_edge(u, v, k, certain)

    def remove_edge(self, u, v, label, certain):
        if certain:
            self.certain_edges.remove((u, v, label))
        else:
            self.uncertain_edges.remove((u, v, label))

    def remove_edges_from(self, edges, certain):
        for (u, v, k) in edges:
            self.remove_edge(u, v, k, certain)

    def add_node(self, node, certain):
        if certain:
            self.certain_vertices.add(node)
        else:
            self.uncertain_vertices.add(node)

    def add_vertices_from(self, nodes, certain):
        for node in nodes:
            self.add_node(node, certain)

    def remove_node(self, node, certain):
        if certain:
            self.certain_vertices.remove(node)
        else:
            self.uncertain_vertices.remove(node)

    def remove_nodes_from(self, nodes, certain):
        for node in nodes:
            self.remove_node(node, certain)

    def __repr__(self) -> str:
        repr = "Evidence(CertainEdges={}, CertainVertices={}," +\
            " UncertainEdges={}, UncertainVertices={})"
        return repr.format(
            str(self.certain_edges),
            str(self.certain_vertices),
            str(self.uncertain_edges),
            str(self.uncertain_vertices)
        )

    def __str__(self) -> str:
        """Returns human readable version of the evidence you've provided."""
        def evidence_str(base_str, edges, vertices):
            if edges == []:
                base_str += '   Edges = []\n'
            else:
                base_str += '   Edges = [\n'
                for edge in edges:
                    base_str += '     %s,\n' % (str(edge))
                base_str += '   ]\n'

            if vertices == set():
                base_str += '   Vertices = {}\n'
            else:
                base_str += '   Vertices = {\n'
                for vertex in vertices:
                    base_str += "     '%s',\n" % (str(vertex))
                base_str += '   }\n\n'
            return base_str

        base_str = 'The evidence you have given is as follows:\n'
        base_str += ' Evidence you are certain of:\n'
        base_str = evidence_str(
            base_str,
            self.certain_edges,
            self.certain_vertices
        )
        base_str += ' Evidence you are uncertain of:\n'
        base_str = evidence_str(
            base_str,
            self.uncertain_edges,
            self.uncertain_vertices
        )
        return base_str

    def __update_path_list(self):
        def remove_paths(paths_list, paths_to_remove):
            for path in paths_to_remove:
                paths_list.remove(path)

        # Certain Evidence
        certain_path_list = self.__graph.path_list.copy()
        # Uncertain Evidence
        uncertain_path_list = self.__graph.path_list.copy()

        if self.certain_edges or self.certain_vertices:
            # Apply certain Edges
            paths_to_remove = []
            for edge in self.certain_edges:
                for path in certain_path_list:
                    if edge not in path:
                        paths_to_remove.append(path)

            remove_paths(certain_path_list, paths_to_remove)

            # Apply certain vertices
            paths_to_remove = []
            for vertex in self.certain_vertices:
                for path in certain_path_list:
                    vertex_in_path = False
                    for (u, v, _) in path:
                        if vertex == u or vertex == v:
                            vertex_in_path = True
                            break

                    if not vertex_in_path:
                        paths_to_remove.append(path)

            remove_paths(certain_path_list, paths_to_remove)

        if self.uncertain_edges or self.uncertain_vertices:
            # Apply uncertain Edges
            paths_to_include = []
            for edge in self.uncertain_edges:
                for path in uncertain_path_list:
                    if edge in path:
                        paths_to_include.append(path)

            # Apply uncertain Vertices
            for vertex in self.uncertain_vertices:
                for path in uncertain_path_list:
                    vertex_in_path = False
                    for (u, v, _) in path:
                        if vertex == u or vertex == v:
                            vertex_in_path = True
                            break

                    if vertex_in_path:
                        paths_to_include.append(path)

            uncertain_path_list = paths_to_include

        # Take paths that are found in both certain and uncertain lists
        if not self.uncertain_edges and not self.uncertain_vertices:
            self.path_list = certain_path_list
        elif not self.certain_edges and not self.certain_vertices:
            self.path_list = uncertain_path_list
        else:
            new_path_list = []
            for path in certain_path_list:
                if path in uncertain_path_list:
                    new_path_list.append(path)
            self.path_list = new_path_list

    def __update_edges_and_vertices(self):
        edges = set()
        vertices = set()

        for path in self.path_list:
            for (u, v, k) in path:
                edges.add((u, v, k))
                vertices.add(u)
                vertices.add(v)

        self.edges = edges
        self.vertices = vertices

    def __filter_edge(self, u, v, k) -> bool:
        if (u, v, k) in self.edges:
            return True
        else:
            return False

    def __filter_node(self, n) -> bool:
        if n in self.vertices:
            return True
        else:
            return False

    def __propagate_reduced_graph_probabilities(self, graph) -> None:
        sink = graph.sink_node
        root = graph.root_node
        graph.nodes[sink]['emphasis'] = 1
        node_set = set([sink])

        while node_set != {root}:
            try:
                v_node = node_set.pop()
            except KeyError:
                break

            for u_node, edges in graph.pred[v_node].items():
                for edge_label, data in edges.items():
                    edge = (u_node, v_node, edge_label)
                    emph = graph.nodes[v_node]['emphasis']
                    prob = data['probability']
                    graph.edges[edge]['potential'] = \
                        prob * emph
                successors = graph.succ[u_node]

                try:
                    emphasis = 0
                    for _, edges in successors.items():
                        for edge_label, data in edges.items():
                            emphasis += data['potential']

                    graph.nodes[u_node]['emphasis'] = emphasis
                    node_set.add(u_node)
                except KeyError:
                    pass

        for (u, v, k) in list(graph.edges):
            edge_potential = graph.edges[(u, v, k)]['potential']
            pred_node_emphasis = graph.nodes[u]['emphasis']
            probability = edge_potential / pred_node_emphasis
            graph.edges[(u, v, k)]['probability'] = probability

    def __create_reduced_graph(self) -> ChainEventGraph:
        self.__update_path_list()
        self.__update_edges_and_vertices()
        subgraph = ChainEventGraph(
            nx.subgraph_view(
                G=self.__graph,
                filter_node=self.__filter_node,
                filter_edge=self.__filter_edge
            )
        ).copy()

        self.__propagate_reduced_graph_probabilities(subgraph)

        return subgraph



"""
CEGpy
=====

CEGpy is a python package for the creation, manipulation, and study of
Chain Event Graphs.

See documentation of more information.

"""
from .utilities.util import *
from .trees.event import *
from .trees.staged import *
from .graphs.ceg import *
import logging
# from graphs import ctceg


logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger('cegpy')



from collections import defaultdict
import numpy as np
import pydotplus as pdp
import logging
from ..utilities.util import Util
from IPython.display import Image
from IPython import get_ipython
import pandas as pd
import textwrap
import networkx as nx
# create logger object for this module
logger = logging.getLogger('cegpy.event_tree')


class EventTree(nx.MultiDiGraph):
    """
    Class for event trees.

    This class extends the networkx DiGraph class to allow the creation
    of event trees from data provided in a pandas dataframe.

    A DiGraph stores nodes and edges with optional data, or attributes.

    DiGraphs hold directed edges.  Self loops are allowed but multiple
    (parallel) edges are not.

    Nodes can be arbitrary (hashable) Python objects with optional
    key/value attributes. By convention `None` is not used as a node.

    Edges are represented as links between nodes with optional
    key/value attributes.

    Parameters
    ----------
    dataframe : Pandas dataframe (required)
        Dataframe containing variables as column headers, with event
        name strings in each cell. These event names will be used to
        create the edges of the event tree. Counts of each event will
        be extracted and attached to each edge.

    sampling_zero_paths: list of tuples containing paths to sampling
        zeros.
        Format is as follows: \
            [('edge_1',), ('edge_1', 'edge_2'), ...]

    incoming_graph_data : input graph (optional, default: None)
        Data to initialize graph.  If None (default) an empty
        graph is created.  The data can be an edge list, or any
        NetworkX graph object.  If the corresponding optional Python
        packages are installed the data can also be a NumPy matrix
        or 2d ndarray, a SciPy sparse matrix, or a PyGraphviz graph.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to graph as key=value pairs.

    See Also
    --------
    StagedTrees \n
    ChainEventGraph

    Examples
    --------
    >>> G = nx.Graph()  # or DiGraph, MultiGraph, MultiDiGraph, etc
    >>> G = nx.Graph(name="my graph")
    >>> e = [(1, 2), (2, 3), (3, 4)]  # list of edges
    >>> G = nx.Graph(e)

    Arbitrary graph attribute pairs (key=value) may be assigned

    >>> G = nx.Graph(e, day="Friday")
    >>> G.graph
    {'day': 'Friday'}

    """
    def __init__(self, dataframe, sampling_zero_paths=None,
                 incoming_graph_data=None, **attr) -> None:
        """Initialize an event tree graph with edges, name, or graph attributes.
        This class extends the networkx DiGraph class to allow the creation
        of event trees from data provided in a pandas dataframe.

        Parameters
        ----------
        dataframe : Pandas dataframe (required)
            Dataframe containing variables as column headers, with event
            name strings in each cell. These event names will be used to
            create the edges of the event tree. Counts of each event will
            be extracted and attached to each edge.

        sampling_zero_paths: list of tuples containing paths to sampling
            zeros.
            Format is as follows: \
                [('edge_1',), ('edge_1', 'edge_2'), ...]

        incoming_graph_data : input graph (optional, default: None)
            Data to initialize graph.  If None (default) an empty
            graph is created.  The data can be an edge list, or any
            NetworkX graph object.  If the corresponding optional Python
            packages are installed the data can also be a NumPy matrix
            or 2d ndarray, a SciPy sparse matrix, or a PyGraphviz graph.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to graph as key=value pairs.

        See Also
        --------
        convert

        Examples
        --------
        >>> G = nx.Graph()  # or DiGraph, MultiGraph, MultiDiGraph, etc
        >>> G = nx.Graph(name="my graph")
        >>> e = [(1, 2), (2, 3), (3, 4)]  # list of edges
        >>> G = nx.Graph(e)

        Arbitrary graph attribute pairs (key=value) may be assigned

        >>> G = nx.Graph(e, day="Friday")
        >>> G.graph
        {'day': 'Friday'}

        """
        logger.info('Initialising')
        # Initialise Networkx DiGraph class
        super().__init__(incoming_graph_data, **attr)

        self.sampling_zeros = sampling_zero_paths

        # Paths sorted alphabetically in order of length
        self._sorted_paths = defaultdict(int)

        # pandas dataframe passed via parameters
        self._dataframe = dataframe
        # Format of event_tree dict:

        self.__construct_event_tree()
        logger.info('Initialisation complete!')

    @property
    def root(self) -> str:
        """Root node of the event tree.
        Currently hard coded to 's0'"""
        return 's0'

    @property
    def variables(self) -> list:
        """The column headers of the dataset"""
        vars = list(self._dataframe.columns)
        logger.info('Variables extracted from dataframe were:')
        logger.info(vars)
        return vars

    @property
    def sampling_zeros(self):
        try:
            if self._sampling_zero_paths is None:
                logger.info("EventTree.sampling_zero_paths \
                        has not been set.")
            return self._sampling_zero_paths
        except AttributeError:
            logger.info("something has gone seriously wrong.")

    @sampling_zeros.setter
    def sampling_zeros(self, sz_paths):
        """Use this function to set the sampling zero paths.
        If different to previous value, will re-generate the event tree."""
        if sz_paths is None:
            self._sampling_zero_paths = None
        else:
            # checkes if the user has inputted sz paths correctly
            sz_paths = self.__check_sampling_zero_paths_param(sz_paths)

            if sz_paths:
                self._sampling_zero_paths = sz_paths
            else:
                error_str = "Parameter 'sampling_zero_paths' not in expected format. \
                             Should be a list of tuples like so:\n \
                             [('edge_1',), ('edge_1', 'edge_2'), ...]"
                if logger.getEffectiveLevel() is logging.DEBUG:
                    logger.debug(error_str)
                    raise ValueError(error_str)

    @property
    def situations(self) -> list:
        """Returns list of event tree situations.
        (non-leaf nodes)"""
        return [
            node for node, out_degree in self.out_degree
            if out_degree != 0
        ]

    @property
    def leaves(self) -> list:
        """Returns leaves of the event tree."""
        # if not already generated, create self.leaves
        return [
            node for node, out_degree in self.out_degree
            if out_degree == 0
        ]

    @property
    def edge_counts(self) -> dict:
        '''list of counts along edges. Indexed same as edges and edge_labels'''
        return nx.get_edge_attributes(self, 'count')

    @property
    def categories_per_variable(self) -> dict:
        '''list of number of unique categories/levels for each variable
        (a column in the df)'''
        def display_nan_warning():
            logger.warning(
                textwrap.dedent(
                    """   --- NaNs found in the dataframe!! ---
                    cegpy assumes that NaNs are either structural zeros or
                    structural missing values.
                    Any non-structural missing values must be dealt with
                    prior to providing the dataset to any of the cegpy
                    functions. Any non-structural zeros should be explicitly
                    added into the cegpy objects.
                    --- See documentation for more information. ---"""
                )
            )

        categories_to_ignore = {"N/A", "NA", "n/a", "na", "NAN", "nan"}
        self._catagories_per_variable = {}
        nans_filtered = False

        for var in self.variables:
            categories = set(self._dataframe[var].unique().tolist())
            # remove nan with pandas
            pd_filtered_categories = {x for x in categories if pd.notna(x)}
            if pd_filtered_categories != categories:
                nans_filtered = True

            # remove any string nans that might have made it in.
            filtered_cats = pd_filtered_categories - categories_to_ignore
            if pd_filtered_categories != filtered_cats:
                nans_filtered = True

            self._catagories_per_variable[var] = len(filtered_cats)

        if nans_filtered:
            display_nan_warning()

        return self._catagories_per_variable

    def _generate_pdp_graph(self):
        node_list = list(self)
        graph = pdp.Dot(graph_type='digraph', rankdir='LR')
        for edge, count in self.edge_counts.items():
            edge_details = edge[2] + '\n' + str(count)

            graph.add_edge(
                pdp.Edge(
                    edge[0],
                    edge[1],
                    label=edge_details,
                    labelfontcolor="#009933",
                    fontsize="10.0",
                    color="black"
                )
            )

        for node in node_list:
            try:
                fill_colour = self.nodes[node]['colour']
            except KeyError:
                fill_colour = 'lightgrey'
            label = "<" + node[0] + "<SUB>" + node[1:] + "</SUB>" + ">"
            graph.add_node(
                pdp.Node(
                    name=node,
                    label=label,
                    style="filled",
                    fillcolor=fill_colour))
        return graph

    def create_figure(self, filename):
        """Draws the event tree for the process described by the dataset,
        and saves it to "<filename>.filetype". Supports any filetype that
        graphviz supports. e.g: "event_tree.png" or "event_tree.svg" etc.
        """
        filename, filetype = Util.generate_filename_and_mkdir(filename)
        logger.info("--- generating graph ---")
        graph = self._generate_pdp_graph()
        logger.info("--- writing " + filetype + " file ---")
        graph.write(str(filename), format=filetype)

        if get_ipython() is None:
            return None
        else:
            logger.info("--- Exporting graph to notebook ---")
            return Image(graph.create_png())

    def __create_unsorted_paths_dict(self) -> defaultdict:
        """Creates and populates a dictionary of all paths provided in the dataframe,
        in the order in which they are given."""
        unsorted_paths = defaultdict(int)

        for variable_number in range(0, len(self.variables)):
            dataframe_upto_variable = self._dataframe.loc[
                :, self.variables[0:variable_number+1]]

            for row in dataframe_upto_variable.itertuples():
                row = row[1:]
                new_row = [edge_label for edge_label in row if
                           edge_label != np.nan and
                           str(edge_label) != 'NaN' and
                           str(edge_label) != 'nan' and
                           edge_label != '']
                new_row = tuple(new_row)

                # checking if the last edge label in row was nan. That would
                # result in double counting nan must be identified as string
                if (row[-1] != np.nan and str(row[-1]) != 'NaN' and
                   str(row[-1]) != 'nan' and row[-1] != ''):
                    unsorted_paths[new_row] += 1

        return unsorted_paths

    def __create_path_dict_entries(self):
        '''Create path dict entries for each path, including the
        sampling zero paths if any.
        Each path is an ordered sequence of edge labels starting
        from the root.
        The keys in the dict are ordered alphabetically.
        Also calls the method self._sampling_zeros to ensure
        manually added path format is correct.
        Added functionality to remove NaN/null edge labels
        assuming they are structural zeroes'''
        unsorted_paths = self.__create_unsorted_paths_dict()

        if self.sampling_zeros is not None:
            unsorted_paths = Util.create_sampling_zeros(
                self.sampling_zeros, unsorted_paths)

        depth = len(max(list(unsorted_paths.keys()), key=len))
        keys_of_list = list(unsorted_paths.keys())
        sorted_keys = []
        for deep in range(0, depth + 1):
            unsorted_mini_list = [key for key in keys_of_list if
                                  len(key) == deep]
            sorted_keys = sorted_keys + sorted(unsorted_mini_list)

        for key in sorted_keys:
            self._sorted_paths[key] = unsorted_paths[key]

        node_list = self.__create_node_list_from_paths(self._sorted_paths)
        self.add_nodes_from(node_list)

    def __check_sampling_zero_paths_param(self, sampling_zero_paths) -> list:
        """Check param 'sampling_zero_paths' is in the correct format"""
        for tup in sampling_zero_paths:
            if not isinstance(tup, tuple):
                return None
            else:
                if not Util.check_tuple_contains_strings(tup):
                    return None

        return sampling_zero_paths

    def __create_node_list_from_paths(self, paths) -> list:
        """Creates list of all nodes: includes root, situations, leaves"""
        node_list = [self.root]

        for vertex_number, _ in enumerate(list(paths.keys()), start=1):
            node_list.append('s%d' % vertex_number)

        return node_list

    def __construct_event_tree(self):
        """Constructs event_tree DiGraph.
        Takes the paths, and adds all the nodes and edges to the Graph"""

        logger.info('Starting construction of event tree')
        self.__create_path_dict_entries()
        # Taking a list of a networkx graph object (self) provides a list
        # of all the nodes
        node_list = list(self)

        # Work through the sorted paths list to build the event tree.
        edge_labels_list = ['root']
        for path, count in list(self._sorted_paths.items()):
            path = list(path)
            edge_labels_list.append(path)
            if path[:-1] in edge_labels_list:
                path_edge_comes_from = edge_labels_list.index(path[:-1])
                self.add_edge(
                    u_for_edge=node_list[path_edge_comes_from],
                    v_for_edge=node_list[edge_labels_list.index(path)],
                    key=path[-1],
                    count=count
                )
            else:
                self.add_edge(
                    u_for_edge=node_list[0],
                    v_for_edge=node_list[edge_labels_list.index(path)],
                    key=path[-1],
                    count=count
                )



import os
import colorutils
import math
from collections import defaultdict
from pathlib import Path
from colorutils import Color as Colour
from datetime import datetime

ST_OUTPUT = {
    "Merged Situations": [],
    "Loglikelihood": float,
    "Mean Posterior Probabilities": [],
}


class Util:
    @staticmethod
    def check_list_contains_strings(str_list) -> bool:
        """Ensure that a list only contains strings"""
        for tup in str_list:
            if not isinstance(tup, str):
                return False
        return True

    @staticmethod
    def check_tuple_contains_strings(tup) -> bool:
        """Check each element of the tuple to ensure it is a string"""
        if isinstance(tup, tuple):
            for elem in tup:
                if not isinstance(elem, str):
                    return False
            return True
        else:
            return False

    @staticmethod
    def generate_colours(number) -> list:
        """Generates a list of hex colour strings that are evenly spaced
        around the colour spectrum"""
        hex_colours = []
        low, start, high = 100, 150, 255
        starts = [
            Colour((start, low, low)),
            Colour((low, start, low)),
            Colour((low, low, start))
        ]
        ends = [
            Colour((high, low, low)),
            Colour((low, high, low)),
            Colour((low, low, high))
        ]
        prim_colours = math.ceil(number / 2)
        hex_colours += Util.generate_colour_run(prim_colours, starts, ends)
        starts = [
            Colour((start, start, low)),
            Colour((low, start, start)),
            Colour((start, low, start))
        ]
        ends = [
            Colour((high, high, low)),
            Colour((low, high, high)),
            Colour((high, low, high))
        ]
        sec_colours = math.floor(number / 2)
        hex_colours += Util.generate_colour_run(sec_colours, starts, ends)

        return hex_colours

    @staticmethod
    def generate_colour_run(number, starts, ends) -> list:
        split = [0, 0, 0]
        for i in range(number):
            split[i % len(split)] += 1

        jumps = []
        for val in split:
            jumps.append(max(val - 1, 0))

        colours = []

        for idx in range(len(ends)):
            if jumps[idx] > 0:
                colours += colorutils.color_run(
                    starts[idx], ends[idx], jumps[idx], to_color=True)
            else:
                colours.append(ends[idx])
        hex_colours = []
        for colour in colours:
            hex_colours.append(colour.hex)

        return hex_colours

    @staticmethod
    def generate_filename_and_mkdir(filename) -> str:
        if filename is not Path:
            filename = Path(filename)

        filetype = filename.suffix.strip('.')
        if filename.suffix == filetype:
            filename.joinpath('.png')
            filetype = 'png'

        os.makedirs(filename.parent, exist_ok=True)
        return filename, filetype

    @staticmethod
    def create_relative_path(relative_path) -> Path:
        """Creates pathlib object from a relative path such as:
        '/output/image.png'"""
        return Path(__file__).resolve().parent.joinpath(relative_path)

    @staticmethod
    def create_sampling_zeros(sampling_zero_paths, path_dict) -> defaultdict:
        '''The list of paths to zero must only contain tuples.
        Each tuple is a sampling zero path that needs to be added.
        If multiple edges along a path need to be added, they must be
        added in order. i.e. path[:-1] should already be a key in self.paths
        Eg suppose edge 'eat' already exists as an edge emanating from the
        root.
        To add paths ('eat', 'sleep'), ('eat', 'sleep', 'repeat'),
        the sampling zero parameter should be added as:
        et = ceg({'dataframe': df, 'sampling_zero_paths':
        [('eat', 'sleep',),('eat', 'sleep','repeat',)]})
        '''
        for path in sampling_zero_paths:
            if (path[:-1] in list(path_dict.keys())) or len(path) == 1:
                path_dict[path] = 0
            else:
                raise ValueError("The path up to it's last edge should be \
                    added first. Ensure the tuple ends with a comma.")

        return path_dict

    @staticmethod
    def create_path(filename, add_time=False, filetype='png'):
        dt_string = ''
        if add_time:
            now = datetime.now()
            dt_string = now.strftime("__%d-%m-%Y_%H-%M-%S")
        fig_path = Path(__file__).resolve().parent\
            .parent.parent.parent.joinpath(
                filename + dt_string + '.' + filetype)
        return fig_path

    @staticmethod
    def flatten_list_of_lists(list_of_lists) -> list:
        flat_list = []
        for sublist in list_of_lists:
            flat_list = flat_list + sublist
        return flat_list
